# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from enum import IntEnum

from gravitino.exceptions.base import (
    ConnectionFailedException,
    RESTException,
    IllegalArgumentException,
    NotFoundException,
    InternalError,
    AlreadyExistsException,
    NotEmptyException,
    UnsupportedOperationException,
    ForbiddenException,
    NotInUseException,
    InUseException,
)


class ErrorConstants(IntEnum):
    """Constants representing error codes for responses."""

    # Error codes for REST responses error.
    REST_ERROR_CODE = 1000

    # Error codes for illegal arguments.
    ILLEGAL_ARGUMENTS_CODE = 1001

    # Error codes for internal errors.
    INTERNAL_ERROR_CODE = 1002

    # Error codes for not found.
    NOT_FOUND_CODE = 1003

    # Error codes for already exists.
    ALREADY_EXISTS_CODE = 1004

    # Error codes for non empty.
    NON_EMPTY_CODE = 1005

    # Error codes for unsupported operation.
    UNSUPPORTED_OPERATION_CODE = 1006

    # Error codes for connect to catalog failed.
    CONNECTION_FAILED_CODE = 1007

    # Error codes for forbidden operation.
    FORBIDDEN_CODE = 1008

    # Error codes for operation on a no in use entity.
    NOT_IN_USE_CODE = 1009

    # Error codes for drop an in use entity.
    IN_USE_CODE = 1010

    # Error codes for invalid state.
    UNKNOWN_ERROR_CODE = 1100


EXCEPTION_MAPPING = {
    RESTException: ErrorConstants.REST_ERROR_CODE,
    IllegalArgumentException: ErrorConstants.ILLEGAL_ARGUMENTS_CODE,
    InternalError: ErrorConstants.INTERNAL_ERROR_CODE,
    NotFoundException: ErrorConstants.NOT_FOUND_CODE,
    AlreadyExistsException: ErrorConstants.ALREADY_EXISTS_CODE,
    NotEmptyException: ErrorConstants.NON_EMPTY_CODE,
    UnsupportedOperationException: ErrorConstants.UNSUPPORTED_OPERATION_CODE,
    ConnectionFailedException: ErrorConstants.CONNECTION_FAILED_CODE,
    ForbiddenException: ErrorConstants.FORBIDDEN_CODE,
    NotInUseException: ErrorConstants.NOT_IN_USE_CODE,
    InUseException: ErrorConstants.IN_USE_CODE,
}

ERROR_CODE_MAPPING = {v: k for k, v in EXCEPTION_MAPPING.items()}



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import Optional, cast

from gravitino.api.audit import Audit
from gravitino.api.rel.column import Column
from gravitino.api.rel.expressions.distributions.distribution import Distribution
from gravitino.api.rel.expressions.sorts.sort_order import SortOrder
from gravitino.api.rel.expressions.transforms.transform import Transform
from gravitino.api.rel.indexes.index import Index
from gravitino.api.rel.partitions.partition import Partition
from gravitino.api.rel.table import Table
from gravitino.client.generic_column import GenericColumn
from gravitino.dto.rel.partitions.partition_dto import PartitionDTO
from gravitino.dto.rel.table_dto import TableDTO
from gravitino.dto.requests.add_partitions_request import AddPartitionsRequest
from gravitino.dto.responses.drop_response import DropResponse
from gravitino.dto.responses.partition_list_response import PartitionListResponse
from gravitino.dto.responses.partition_name_list_response import (
    PartitionNameListResponse,
)
from gravitino.dto.responses.partition_response import PartitionResponse
from gravitino.dto.util.dto_converters import DTOConverters
from gravitino.exceptions.handlers.partition_error_handler import (
    PARTITION_ERROR_HANDLER,
)
from gravitino.namespace import Namespace
from gravitino.rest.rest_utils import encode_string
from gravitino.utils import HTTPClient


class RelationalTable(Table):
    """Represents a relational table."""

    def __init__(
        self, namespace: Namespace, table_dto: TableDTO, rest_client: HTTPClient
    ):
        self._namespace = namespace
        self._table = cast(Table, DTOConverters.from_dto(table_dto))
        self._rest_client = rest_client

    def name(self) -> str:
        return self._table.name()

    def columns(self) -> list[Column]:
        metalake, catalog, schema = self._namespace.levels()
        return [
            cast(
                Column,
                GenericColumn(
                    c, self._rest_client, metalake, catalog, schema, self._table.name()
                ),
            )
            for c in self._table.columns()
        ]

    def partitioning(self) -> list[Transform]:
        return self._table.partitioning()

    def sort_order(self) -> list[SortOrder]:
        return self._table.sort_order()

    def distribution(self) -> Distribution:
        return self._table.distribution()

    def index(self) -> list[Index]:
        return self._table.index()

    def comment(self) -> Optional[str]:
        return self._table.comment()

    def properties(self) -> dict[str, str]:
        return self._table.properties()

    def audit_info(self) -> Audit:
        return self._table.audit_info()

    def _get_partition_request_path(self) -> str:
        """Get the partition request path.

        Returns:
            str: The partition request path.
        """

        return (
            f"api/metalakes/{encode_string(self._namespace.level(0))}"
            f"/catalogs/{encode_string(self._namespace.level(1))}"
            f"/schemas/{encode_string(self._namespace.level(2))}"
            f"/tables/{encode_string(self._table.name())}"
            "/partitions"
        )

    def list_partition_names(self) -> list[str]:
        """Get the partition names of the table.

        Returns:
            list[str]: The partition names of the table.
        """

        resp = self._rest_client.get(
            endpoint=self._get_partition_request_path(),
            error_handler=PARTITION_ERROR_HANDLER,
        )
        partition_name_list_resp = PartitionNameListResponse.from_json(
            resp.body, infer_missing=True
        )
        partition_name_list_resp.validate()

        return partition_name_list_resp.partition_names()

    def list_partitions(self) -> list[Partition]:
        """Get the partitions of the table.

        Returns:
            list[Partition]: The partitions of the table.
        """

        params = {"details": "true"}
        resp = self._rest_client.get(
            endpoint=self._get_partition_request_path(),
            params=params,
            error_handler=PARTITION_ERROR_HANDLER,
        )
        partition_list_resp = PartitionListResponse.from_json(
            resp.body, infer_missing=True
        )
        partition_list_resp.validate()

        return partition_list_resp.get_partitions()

    def get_partition(self, partition_name: str) -> Partition:
        """Returns the partition with the given name.

        Args:
            partition_name (str): the name of the partition

        Returns:
            Partition: the partition with the given name

        Raises:
            NoSuchPartitionException:
                if the partition does not exist, throws this exception.
        """

        resp = self._rest_client.get(
            endpoint=f"{self._get_partition_request_path()}/{encode_string(partition_name)}",
            error_handler=PARTITION_ERROR_HANDLER,
        )
        partition_resp = PartitionResponse.from_json(resp.body, infer_missing=True)
        partition_resp.validate()

        return partition_resp.get_partition()

    def drop_partition(self, partition_name: str) -> bool:
        """Drops the partition with the given name.

        Args:
            partition_name (str): The name of the partition.

        Returns:
            bool: `True` if the partition is dropped, `False` if the partition does not exist.
        """
        resp = self._rest_client.delete(
            endpoint=f"{self._get_partition_request_path()}/{encode_string(partition_name)}",
            error_handler=PARTITION_ERROR_HANDLER,
        )
        drop_resp = DropResponse.from_json(resp.body, infer_missing=True)
        drop_resp.validate()

        return drop_resp.dropped()

    def add_partition(self, partition: Partition) -> Partition:
        """Adds a partition to the table.

        Args:
            partition (Partition): The partition to add.

        Returns:
            Partition: The added partition.

        Raises:
            PartitionAlreadyExistsException:
                if the partition already exists, throws this exception.
        """

        req = AddPartitionsRequest(
            [cast(PartitionDTO, DTOConverters.to_dto(partition))]
        )
        req.validate()

        resp = self._rest_client.post(
            endpoint=self._get_partition_request_path(),
            json=req,
            error_handler=PARTITION_ERROR_HANDLER,
        )
        partition_list_resp = PartitionListResponse.from_json(
            resp.body, infer_missing=True
        )
        partition_list_resp.validate()
        return partition_list_resp.get_partitions()[0]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import ABC, abstractmethod
from typing import List, Dict
from .job_handle import JobHandle
from .job_template import JobTemplate
from .job_template_change import JobTemplateChange


class SupportsJobs(ABC):
    """
    Interface for job management operations. This interface will be mixed with GravitinoClient to
    provide the ability to manage job templates and jobs within the Gravitino system.
    """

    @abstractmethod
    def list_job_templates(self) -> List[JobTemplate]:
        """
        Lists all the registered job templates in Gravitino.

        Returns:
            List of job templates.
        """
        pass

    @abstractmethod
    def register_job_template(self, job_template) -> None:
        """
        Register a job template with the specified job template to Gravitino. The registered
        job will be maintained in Gravitino, allowing it to be executed later.

        Args:
            job_template: The job template to register.

        Raises:
            JobTemplateAlreadyExists: If a job template with the same name already exists.
        """
        pass

    @abstractmethod
    def get_job_template(self, job_template_name: str) -> JobTemplate:
        """
        Retrieves a job template by its name.

        Args:
            job_template_name: The name of the job template to retrieve.

        Returns:
            The job template if found, otherwise raises an exception.

        Raises:
            NoSuchJobTemplateException: If no job template with the specified name exists.
        """
        pass

    @abstractmethod
    def delete_job_template(self, job_template_name: str) -> bool:
        """
        Deletes a job template by its name. This will remove the job template from Gravitino, and it
        will no longer be available for execution. Only when all the jobs associated with this job
        template are completed, failed or cancelled, the job template can be deleted successfully,
        otherwise it will return false.

        The deletion of a job template will also delete all the jobs associated with this template.

        Args:
            job_template_name: The name of the job template to delete.

        Returns:
            bool: True if the job template was deleted successfully, False if the job template
            does not exist.

        Raises:
            InUseException: If there are still queued or started jobs associated with this job template.
        """
        pass

    def alter_job_template(
        self, job_template_name: str, *changes: JobTemplateChange
    ) -> JobTemplate:
        """
        Alters a job template by applying the specified changes. This allows for modifying the
        properties of an existing job template, such as its name, description, or parameters.

        Args:
            job_template_name: The name of the job template.
            changes: The changes to apply to the job template.

        Raises:
            NoSuchJobTemplateException: If the job template does not exist.
            IllegalArgumentException: If any of the changes cannot be applied to the job template.

        Returns:
            The altered job template.
        """
        pass

    @abstractmethod
    def list_jobs(self, job_template_name: str = None) -> List[JobHandle]:
        """
        Lists all the jobs in Gravitino. If a job template name is provided, it will filter the jobs
        by that template.

        Args:
            job_template_name: Optional; if provided, only jobs associated with this template will be listed.

        Returns:
            List of JobHandle objects representing the jobs.
        """
        pass

    @abstractmethod
    def run_job(self, job_template_name: str, job_conf: Dict[str, str]) -> JobHandle:
        """
        Run a job with the template name and configuration. The jobConf is a map of key-value
        contains the variables that will be used to replace the templated parameters in the job
        template.

        Args:
            job_template_name: The name of the job template to run.
            job_conf: A dictionary containing the job configuration parameters.

        Returns:
            JobHandle: A handle to the running job, which can be used to monitor or manage the job.

        Raises:
            NoSuchJobTemplateException: If no job template with the specified name exists.
        """
        pass

    @abstractmethod
    def get_job(self, job_id: str) -> JobHandle:
        """
        Retrieves a job by its ID.

        Args:
            job_id: The ID of the job to retrieve.

        Returns:
            JobHandle: The handle representing the job with the specified ID.

        Raises:
            NoSuchJobException: If no job with the specified ID exists.
        """
        pass

    @abstractmethod
    def cancel_job(self, job_id: str) -> JobHandle:
        """
        Cancels a running job by its ID. This operation will attempt to cancel the job if it is
        still running. This method will return immediately, user could use the job handle to
        check the status of the job after invoking this method.

        Args:
            job_id: The ID of the job to cancel.

        Raises:
            NoSuchJobException: If no job with the specified ID exists.
        """
        pass



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import logging
import configparser
import os
import os.path

from gravitino.auth.auth_data_provider import AuthDataProvider
from gravitino.client.gravitino_metalake import GravitinoMetalake
from gravitino.client.gravitino_version import GravitinoVersion
from gravitino.dto.version_dto import VersionDTO
from gravitino.dto.responses.metalake_response import MetalakeResponse
from gravitino.dto.responses.version_response import VersionResponse
from gravitino.exceptions.handlers.metalake_error_handler import METALAKE_ERROR_HANDLER
from gravitino.exceptions.handlers.rest_error_handler import REST_ERROR_HANDLER
from gravitino.rest.rest_utils import encode_string
from gravitino.utils import HTTPClient
from gravitino.exceptions.base import GravitinoRuntimeException
from gravitino.constants.version import VERSION_INI, Version
from gravitino.name_identifier import NameIdentifier

logger = logging.getLogger(__name__)


class GravitinoClientBase:
    """
    Base class for Gravitino Java client;
    It uses an underlying {@link RESTClient} to send HTTP requests and receive responses from the API.
    """

    _rest_client: HTTPClient
    """The REST client to communicate with the REST server"""

    API_METALAKES_LIST_PATH = "api/metalakes"
    """The REST API path for listing metalakes"""

    API_METALAKES_IDENTIFIER_PATH = f"{API_METALAKES_LIST_PATH}/"
    """The REST API path prefix for load a specific metalake"""

    CLIENT_VERSION_HEADER = "X-Client-Version"
    VERSION_CHECK_DISABLED_ENV = "GRAVITINO_VERSION_CHECK_DISABLED"

    def __init__(
        self,
        uri: str,
        check_version: bool = True,
        auth_data_provider: AuthDataProvider = None,
        request_headers: dict = None,
        client_config: dict = None,
    ):
        new_headers = request_headers.copy() if request_headers else {}
        client_version = self.get_client_version().version()
        if client_version is not None:
            new_headers[self.CLIENT_VERSION_HEADER] = client_version.strip()
        self._rest_client = HTTPClient(
            uri,
            auth_data_provider=auth_data_provider,
            request_headers=new_headers,
            client_config=client_config,
        )
        if check_version and not self._is_version_check_disabled_by_env():
            self.check_version()

    def _is_version_check_disabled_by_env(self) -> bool:
        env_value = os.getenv(self.VERSION_CHECK_DISABLED_ENV)
        return env_value is not None and env_value.strip().lower() == "true"

    def load_metalake(self, name: str) -> GravitinoMetalake:
        """Loads a specific Metalake from the Gravitino API.

        Args:
            name: The name of the Metalake to be loaded.

        Returns:
            A GravitinoMetalake instance representing the loaded Metalake.

        Raises:
            NoSuchMetalakeException If the specified Metalake does not exist.
        """

        self.check_metalake_name(name)
        response = self._rest_client.get(
            GravitinoClientBase.API_METALAKES_IDENTIFIER_PATH + encode_string(name),
            error_handler=METALAKE_ERROR_HANDLER,
        )
        metalake_response = MetalakeResponse.from_json(
            response.body, infer_missing=True
        )
        metalake_response.validate()

        return GravitinoMetalake(metalake_response.metalake(), self._rest_client)

    def check_version(self):
        """Check the compatibility of the client with the target server.

        Raises:
            GravitinoRuntimeException If the client version is greater than the server version.
        """
        server_version = self.get_server_version()
        client_version = self.get_client_version()

        if not client_version.compatible_with_server_version(server_version):
            raise GravitinoRuntimeException(
                "Gravitino does not support the case that "
                "the client-side version is higher than the server-side version."
                f"The client version is {client_version.version()}, and the server version {server_version.version()}"
            )

    def get_client_version(self) -> GravitinoVersion:
        """Retrieves the version of the Gravitino Python Client.

        Returns:
            A GravitinoVersion instance representing the version of the Gravitino Python Client.
        """
        config = configparser.ConfigParser()

        if not os.path.exists(VERSION_INI):
            raise GravitinoRuntimeException(
                f"Failed to get Gravitino version, version file '{VERSION_INI}' does not exist."
            )
        config.read(VERSION_INI)

        version = config["metadata"][Version.VERSION.value]
        compile_date = config["metadata"][Version.COMPILE_DATE.value]
        git_commit = config["metadata"][Version.GIT_COMMIT.value]

        return GravitinoVersion(VersionDTO(version, compile_date, git_commit))

    def get_server_version(self) -> GravitinoVersion:
        """Retrieves the version of the Gravitino API.

        Returns:
            A GravitinoVersion instance representing the version of the Gravitino API.
        """
        resp = self._rest_client.get("api/version", error_handler=REST_ERROR_HANDLER)
        version_response = VersionResponse.from_json(resp.body, infer_missing=True)
        version_response.validate()

        return GravitinoVersion(version_response.version())

    def close(self):
        """Closes the GravitinoClient and releases any underlying resources."""
        if self._rest_client is not None:
            self._rest_client.close()

    def check_metalake_name(self, metalake_name: str):
        identifier = NameIdentifier.parse(metalake_name)
        namespace = identifier.namespace()

        if not namespace:
            raise ValueError(
                f"Metalake namespace must be empty, the input namespace is {namespace}"
            )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import List, Optional

from gravitino.api.rel.expressions.expression import Expression
from gravitino.api.rel.expressions.named_reference import NamedReference
from gravitino.dto.rel.column_dto import ColumnDTO
from gravitino.dto.rel.partition_utils import PartitionUtils
from gravitino.dto.rel.partitioning.partitioning import Partitioning
from gravitino.dto.rel.partitions.range_partition_dto import RangePartitionDTO


class RangePartitioningDTO(Partitioning):
    """Data transfer object representing a range partitioning."""

    def __init__(
        self,
        field_name: List[str],
        assignments: Optional[List[RangePartitionDTO]] = None,
    ):
        self._field_name = field_name
        self._assignments = assignments if assignments is not None else []

    def field_name(self) -> List[str]:
        """Returns the name of the field to partition.

        Returns:
            List[str]: The name of the field to partition.
        """
        return self._field_name

    def assignments(self) -> List[RangePartitionDTO]:
        return self._assignments

    def strategy(self) -> Partitioning.Strategy:
        return self.Strategy.RANGE

    def validate(self, columns: List[ColumnDTO]) -> None:
        PartitionUtils.validate_field_existence(columns, self._field_name)

    def name(self) -> str:
        return self.strategy().name.lower()

    def arguments(self) -> List[Expression]:
        """Returns the arguments of the partitioning strategy.

        Returns:
            List[Expression]: The arguments of the partitioning strategy.
        """
        return [NamedReference.field(field_name=self._field_name)]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import ClassVar, Optional

from gravitino.api.auditable import Auditable
from gravitino.api.metadata_object import MetadataObject
from gravitino.exceptions.base import UnsupportedOperationException


class AssociatedObjects(ABC):
    """The interface of the associated objects of the tag."""

    @abstractmethod
    def count(self) -> int:
        """Get the number of associated objects.

        Returns:
            int: The number of associated objects.
        """
        objects = self.objects()
        return 0 if objects is None else len(objects)

    @abstractmethod
    def objects(self) -> Optional[list[MetadataObject]]:
        """Get the associated objects.

        Returns:
            Optional[list[MetadataObject]]: The list of objects that are associated with this tag..
        """
        pass


class Tag(Auditable):
    """The interface of a tag.

    A tag is a label that can be attached to a catalog, schema, table, fileset, topic,
    or column. It can be used to categorize, classify, or annotate these objects.
    """

    PROPERTY_COLOR: ClassVar[str] = "color"
    """
    A reserved property to specify the color of the tag. The color is a string of hex code that
    represents the color of the tag. The color is used to visually distinguish the tag from other
    tags.
    """

    @abstractmethod
    def name(self) -> str:
        """Get the name of the tag.

        Returns:
            str: The name of the tag.
        """
        raise NotImplementedError()

    @abstractmethod
    def comment(self) -> str:
        """Get the comment of the tag.

        Returns:
            str: The comment of the tag.
        """
        raise NotImplementedError()

    @abstractmethod
    def properties(self) -> dict[str, str]:
        """Get the properties of the tag.

        Returns:
            Dict[str, str]: The properties of the tag.
        """
        raise NotImplementedError()

    @abstractmethod
    def inherited(self) -> Optional[bool]:
        """Check if the tag is inherited from a parent object or not.

        If the tag is inherited, it will return `True`, if it is owned by the object itself, it will return `False`.

        **Note**. The return value is optional, only when the tag is associated with an object, and called from the
        object, the return value will be present. Otherwise, it will be empty.

        Returns:
            Optional[bool]:
                True if the tag is inherited, false if it is owned by the object itself. Empty if the
                tag is not associated with any object.
        """
        raise NotImplementedError()

    def associated_objects(self) -> AssociatedObjects:
        """The associated objects of the tag.

        Raises:
            UnsupportedOperationException: The associated_objects method is not supported.

        Returns:
            AssociatedObjects: The associated objects of the tag.
        """
        raise UnsupportedOperationException(
            "The associated_objects method is not supported."
        )

    class AssociatedObjects(ABC):
        """The interface of the associated objects of the tag."""

        def count(self) -> int:
            """
            Retrieve the number of objects that are associated with this Tag

            Returns:
                int: The number of objects that are associated with this Tag
            """
            return 0 if (s := self.objects()) is None else len(s)

        @abstractmethod
        def objects(self) -> list[MetadataObject]:
            """
            Retrieve the list of objects that are associated with this tag.

            Raises:
                NotImplementedError: if the method is not implemented.

            Returns:
                list[MetadataObject]: The list of objects that are associated with this tag.
            """
            raise NotImplementedError()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import Dict, Optional
from abc import abstractmethod

from gravitino.api.auditable import Auditable


class Model(Auditable):
    """An interface representing an ML model under a schema `Namespace`. A model is a metadata
    object that represents the model artifact in ML. Users can register a model object in Gravitino
    to manage the model metadata. The typical use case is to manage the model in ML lifecycle with a
    unified way in Gravitino, and access the model artifact with a unified identifier. Also, with
    the model registered in Gravitino, users can govern the model with Gravitino's unified audit,
    tag, and role management.

    The difference of Model and tabular data is that the model is schema-free, and the main
    property of the model is the model artifact URL. The difference compared to the fileset is that
    the model is versioned, and the model object contains the version information.
    """

    @abstractmethod
    def name(self) -> str:
        """
        Returns:
            Name of the model object.
        """
        pass

    @abstractmethod
    def comment(self) -> Optional[str]:
        """The comment of the model object. This is the general description of the model object.
        User can still add more detailed information in the model version.

        Returns:
            The comment of the model object. None is returned if no comment is set.
        """
        pass

    def properties(self) -> Dict[str, str]:
        """The properties of the model object. The properties are key-value pairs that can be used
        to store additional information of the model object. The properties are optional.

        Users can still specify the properties in the model version for different information.

        Returns:
            The properties of the model object. An empty dictionary is returned if no properties are set.
        """
        pass

    @abstractmethod
    def latest_version(self) -> int:
        """The latest version of the model object. The latest version is the version number of the
        latest model checkpoint / snapshot that is linked to the registered model.

        Returns:
            The latest version of the model object.
        """
        pass



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import logging
from typing import Dict, List, Optional

from gravitino.api.catalog import Catalog
from gravitino.api.metadata_object import MetadataObject
from gravitino.api.metadata_objects import MetadataObjects
from gravitino.api.function.function import Function
from gravitino.api.function.function_catalog import FunctionCatalog
from gravitino.api.function.function_change import FunctionChange
from gravitino.api.function.function_definition import FunctionDefinition
from gravitino.api.function.function_type import FunctionType
from gravitino.api.schema import Schema
from gravitino.api.schema_change import SchemaChange
from gravitino.api.supports_schemas import SupportsSchemas
from gravitino.client.function_catalog_operations import FunctionCatalogOperations
from gravitino.client.metadata_object_credential_operations import (
    MetadataObjectCredentialOperations,
)
from gravitino.dto.audit_dto import AuditDTO
from gravitino.dto.catalog_dto import CatalogDTO
from gravitino.dto.requests.schema_create_request import SchemaCreateRequest
from gravitino.dto.requests.schema_update_request import SchemaUpdateRequest
from gravitino.dto.requests.schema_updates_request import SchemaUpdatesRequest
from gravitino.dto.responses.drop_response import DropResponse
from gravitino.dto.responses.entity_list_response import EntityListResponse
from gravitino.dto.responses.schema_response import SchemaResponse
from gravitino.exceptions.base import IllegalArgumentException
from gravitino.exceptions.handlers.schema_error_handler import SCHEMA_ERROR_HANDLER
from gravitino.name_identifier import NameIdentifier
from gravitino.namespace import Namespace
from gravitino.rest.rest_utils import encode_string
from gravitino.utils import HTTPClient

logger = logging.getLogger(__name__)


class BaseSchemaCatalog(CatalogDTO, SupportsSchemas, FunctionCatalog):
    """
    BaseSchemaCatalog is the base abstract class for all the catalog with schema. It provides the
    common methods for managing schemas in a catalog. With BaseSchemaCatalog, users can list,
    create, load, alter and drop a schema with specified identifier.
    """

    # The REST client to send the requests.
    rest_client: HTTPClient

    # The namespace of current catalog, which is the metalake name.
    _catalog_namespace: Namespace

    # The metadata object credential operations
    _object_credential_operations: MetadataObjectCredentialOperations

    _function_operations: FunctionCatalogOperations

    def __init__(
        self,
        catalog_namespace: Namespace,
        name: str = None,
        catalog_type: Catalog.Type = Catalog.Type.UNSUPPORTED,
        provider: str = None,
        comment: str = None,
        properties: Dict[str, str] = None,
        audit: AuditDTO = None,
        rest_client: HTTPClient = None,
    ):
        super().__init__(
            _name=name,
            _type=catalog_type,
            _provider=provider,
            _comment=comment,
            _properties=properties,
            _audit=audit,
        )
        self.rest_client = rest_client
        self._catalog_namespace = catalog_namespace

        metadata_object = MetadataObjects.of([name], MetadataObject.Type.CATALOG)
        self._object_credential_operations = MetadataObjectCredentialOperations(
            catalog_namespace.level(0), metadata_object, rest_client
        )
        self._function_operations = FunctionCatalogOperations(
            rest_client, catalog_namespace, self.name()
        )

        self.validate()

    def as_schemas(self):
        return self

    def as_function_catalog(self):
        return self

    def list_schemas(self) -> List[str]:
        """List all the schemas under the given catalog namespace.

        Raises:
            NoSuchCatalogException if the catalog with specified namespace does not exist.

        Returns:
             A list of schema names under the given catalog namespace.
        """
        resp = self.rest_client.get(
            BaseSchemaCatalog.format_schema_request_path(self._schema_namespace()),
            error_handler=SCHEMA_ERROR_HANDLER,
        )
        entity_list_response = EntityListResponse.from_json(
            resp.body, infer_missing=True
        )
        entity_list_response.validate()

        return [ident.name() for ident in entity_list_response.identifiers()]

    def create_schema(
        self,
        schema_name: str = None,
        comment: str = None,
        properties: Dict[str, str] = None,
    ) -> Schema:
        """Create a new schema with specified identifier, comment and metadata.

        Args:
            schema_name: The name of the schema.
            comment: The comment of the schema.
            properties: The properties of the schema.

        Raises:
            NoSuchCatalogException if the catalog with specified namespace does not exist.
            SchemaAlreadyExistsException if the schema with specified identifier already exists.

        Returns:
             The created Schema.
        """
        req = SchemaCreateRequest(encode_string(schema_name), comment, properties)
        req.validate()

        resp = self.rest_client.post(
            BaseSchemaCatalog.format_schema_request_path(self._schema_namespace()),
            json=req,
            error_handler=SCHEMA_ERROR_HANDLER,
        )
        schema_response = SchemaResponse.from_json(resp.body, infer_missing=True)
        schema_response.validate()

        return schema_response.schema()

    def load_schema(self, schema_name: str) -> Schema:
        """Load the schema with specified identifier.

        Args:
            schema_name: The name of the schema.

        Raises:
            NoSuchSchemaException if the schema with specified identifier does not exist.

        Returns:
            The Schema with specified identifier.
        """
        resp = self.rest_client.get(
            BaseSchemaCatalog.format_schema_request_path(self._schema_namespace())
            + "/"
            + encode_string(schema_name),
            error_handler=SCHEMA_ERROR_HANDLER,
        )
        schema_response = SchemaResponse.from_json(resp.body, infer_missing=True)
        schema_response.validate()

        return schema_response.schema()

    def alter_schema(self, schema_name: str, *changes: SchemaChange) -> Schema:
        """Alter the schema with specified identifier by applying the changes.

        Args:
            schema_name: The name of the schema.
            changes: The metadata changes to apply.

        Raises:
            NoSuchSchemaException if the schema with specified identifier does not exist.

        Returns:
            The altered Schema.
        """
        reqs = [
            BaseSchemaCatalog.to_schema_update_request(change) for change in changes
        ]
        updates_request = SchemaUpdatesRequest(reqs)
        updates_request.validate()
        resp = self.rest_client.put(
            BaseSchemaCatalog.format_schema_request_path(self._schema_namespace())
            + "/"
            + encode_string(schema_name),
            updates_request,
            error_handler=SCHEMA_ERROR_HANDLER,
        )
        schema_response = SchemaResponse.from_json(resp.body, infer_missing=True)
        schema_response.validate()
        return schema_response.schema()

    def drop_schema(self, schema_name: str, cascade: bool) -> bool:
        """Drop the schema with specified identifier.

        Args:
            schema_name: The name of the schema.
            cascade: Whether to drop all the tables under the schema.

        Raises:
            NonEmptySchemaException if the schema is not empty and cascade is false.

        Returns:
             true if the schema is dropped successfully, false otherwise.
        """
        params = {"cascade": str(cascade)}
        resp = self.rest_client.delete(
            BaseSchemaCatalog.format_schema_request_path(self._schema_namespace())
            + "/"
            + encode_string(schema_name),
            params=params,
            error_handler=SCHEMA_ERROR_HANDLER,
        )
        drop_resp = DropResponse.from_json(resp.body, infer_missing=True)
        drop_resp.validate()
        return drop_resp.dropped()

    def list_functions(self, namespace: Namespace) -> List[NameIdentifier]:
        return self._function_operations.list_functions(namespace)

    def list_function_infos(self, namespace: Namespace) -> List[Function]:
        return self._function_operations.list_function_infos(namespace)

    def get_function(self, ident: NameIdentifier) -> Function:
        return self._function_operations.get_function(ident)

    def register_function(
        self,
        ident: NameIdentifier,
        comment: Optional[str],
        function_type: FunctionType,
        deterministic: bool,
        definitions: List[FunctionDefinition],
    ) -> Function:
        return self._function_operations.register_function(
            ident,
            comment,
            function_type,
            deterministic,
            definitions,
        )

    def alter_function(
        self, ident: NameIdentifier, *changes: FunctionChange
    ) -> Function:
        return self._function_operations.alter_function(ident, *changes)

    def drop_function(self, ident: NameIdentifier) -> bool:
        return self._function_operations.drop_function(ident)

    def _schema_namespace(self) -> Namespace:
        return Namespace.of(self._catalog_namespace.level(0), self.name())

    @staticmethod
    def format_schema_request_path(ns: Namespace):
        return (
            "api/metalakes/"
            + encode_string(ns.level(0))
            + "/catalogs/"
            + encode_string(ns.level(1))
            + "/schemas"
        )

    @staticmethod
    def to_schema_update_request(change: SchemaChange):
        if isinstance(change, SchemaChange.SetProperty):
            return SchemaUpdateRequest.SetSchemaPropertyRequest(
                change.property(), change.value()
            )
        if isinstance(change, SchemaChange.RemoveProperty):
            return SchemaUpdateRequest.RemoveSchemaPropertyRequest(change.property())
        raise ValueError(f"Unknown change type: {type(change).__name__}")

    def validate(self):
        Namespace.check(
            self._catalog_namespace is not None
            and self._catalog_namespace.length() == 1,
            f"Catalog namespace must be non-null and have 1 level, the input namespace is {self._catalog_namespace}",
        )

        if self.rest_client is None:
            raise IllegalArgumentException("restClient must be set")
        if not self.name() or not self.name().strip():
            raise IllegalArgumentException("name must not be blank")
        if self.type() is None:
            raise IllegalArgumentException("type must not be None")
        if not self.provider() or not self.provider().strip():
            raise IllegalArgumentException("provider must not be blank")
        if self.audit_info() is None:
            raise IllegalArgumentException("audit must not be None")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import ABC
from enum import Enum
from typing import Dict, Optional

from gravitino.api.function.function_resources import FunctionResources


class FunctionImpl(ABC):
    """Base class of function implementations.

    A function implementation must declare its language and optional external resources.
    Concrete implementations are provided by SQLImpl, JavaImpl, and PythonImpl.
    """

    class Language(Enum):
        """Supported implementation languages."""

        SQL = "SQL"
        """SQL implementation."""

        JAVA = "JAVA"
        """Java implementation."""

        PYTHON = "PYTHON"
        """Python implementation."""

    class RuntimeType(Enum):
        """Supported execution runtimes for function implementations."""

        SPARK = "SPARK"
        """Spark runtime."""

        TRINO = "TRINO"
        """Trino runtime."""

        @classmethod
        def from_string(cls, value: str) -> "FunctionImpl.RuntimeType":
            """Parse a runtime value from string.

            Args:
                value: Runtime name.

            Returns:
                Parsed runtime.

            Raises:
                ValueError: If the runtime is not supported.
            """
            if not value or not value.strip():
                raise ValueError("Function runtime must be set")
            normalized = value.strip().upper()
            for runtime in cls:
                if runtime.name == normalized:
                    return runtime
            raise ValueError(f"Unsupported function runtime: {value}")

    def __init__(
        self,
        language: Language,
        runtime: RuntimeType,
        resources: Optional[FunctionResources] = None,
        properties: Optional[Dict[str, str]] = None,
    ):
        """Construct a FunctionImpl.

        Args:
            language: The language of the function implementation.
            runtime: The runtime of the function implementation.
            resources: The resources required by the function implementation.
            properties: The properties of the function implementation.

        Raises:
            ValueError: If language or runtime is not set.
        """
        if language is None:
            raise ValueError("Function implementation language must be set")
        self._language = language

        if runtime is None:
            raise ValueError("Function runtime must be set")
        self._runtime = runtime

        self._resources = resources if resources else FunctionResources.empty()
        self._properties = dict(properties) if properties else {}

    def language(self) -> Language:
        """Returns the implementation language."""
        return self._language

    def runtime(self) -> RuntimeType:
        """Returns the target runtime."""
        return self._runtime

    def resources(self) -> FunctionResources:
        """Returns the external resources required by this implementation."""
        return self._resources

    def properties(self) -> Dict[str, str]:
        """Returns the additional properties of this implementation."""
        return dict(self._properties)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import ABC, abstractmethod
from enum import Enum


class JobHandle(ABC):
    class Status(Enum):
        QUEUED = "queued"
        STARTED = "started"
        FAILED = "failed"
        SUCCEEDED = "succeeded"
        CANCELLING = "cancelling"
        CANCELLED = "cancelled"

        @classmethod
        def job_status_serialize(cls, status: "JobHandle.Status") -> str:
            """Serializes the job status to a string."""
            return status.value.lower()

        @classmethod
        def job_status_deserialize(cls, status: str) -> "JobHandle.Status":
            """Deserializes a string to a job status."""
            for m in cls:
                if m.value.lower() == status.lower():
                    return m
            raise ValueError(f"Unknown job status: {status}")

    @abstractmethod
    def job_template_name(self) -> str:
        pass

    @abstractmethod
    def job_id(self) -> str:
        pass

    @abstractmethod
    def job_status(self) -> Status:
        pass
